# Copyright (c) 2012 Cloudera, Inc. All rights reserved.
import time
from datetime import datetime
from time import mktime
import os
import simplejson as json
import re
import gzip
import zipfile
import tempfile
import util
try:
  from cStringIO import StringIO
except:
  from StringIO import StringIO

from clusterstats.log.streaming.file import LogFile
from clusterstats.log.streaming.event_reader import LogFileEventReader
from clusterstats.log.streaming.event_parser import Log4JEventParser, HueEventParser, GLogEventParser, ScmAgentEventParser, ScmServerEventParser

from clusterstats.log.streaming.file_filter import DirectoryFilter
from clusterstats.log.streaming.file_filter import FilePatternFilter
from clusterstats.log.streaming.file_filter import TimeRangeFilter
import clusterstats.common.chain as chain

import log_search

TEN_SECONDS_IN_MILLIS=10000

HUE = "HUE"
LOG4J = "LOG4J"
GLOG = "GLOG"
SCMAGENT = "SCMAGENT"
SCMSERVER = "SCMSERVER"

PYTHON_LOG_TYPES = set([HUE, SCMAGENT])
VALID_LOG_TYPES = set([LOG4J, GLOG, SCMSERVER]) | PYTHON_LOG_TYPES

EVENT_PARSER_CLASSES = {
  LOG4J: Log4JEventParser,
  HUE: HueEventParser,
  GLOG: GLogEventParser,
  SCMAGENT: ScmAgentEventParser,
  SCMSERVER: ScmServerEventParser
}

def retrieve_log_compressed(end_time,
                           log_path,
                           log_type,
                           total_byte_limit,
                           search_timeout_millis=TEN_SECONDS_IN_MILLIS):
  """
  Retrieves a compressed portions of a plaintext log file. If available,
  total_byte_limit bytes will be collected from this file and other rolled
  files, with the latest event being at end_time. 
  A log type (valid values in VALID_LOG_TYPES) must also be passed. 
  An optional search timeout can be used to stop collecting log data.

  This will return a generator that outputs a gzip file of the log events.
  """
  compressor = LogCompressor(log_path, log_type, end_time, total_byte_limit)

  log_gen = compressor.compress_log(search_timeout_millis)

  return log_gen

def build_gzip_generator_single(log_file_handle, start_offset, end_offset, deadline=None):
  """
  Gets a generator for gzipped content from the given log file handle, start offset, and end offset.
  The file handle should already be open. This method will close it.
  The generator will return gzipped content of the file from start offset 
  to end offset.
  If a deadline is passed, compression will stop immediately if the deadline time
  is hit.
  """

  CHUNK_SIZE = 128*1024
  if start_offset == end_offset:
    return
    
  log_file_handle.seek(start_offset)

  bytes_to_read = end_offset - start_offset
  while bytes_to_read > 0:
    if deadline != None and time.time() > deadline:
      log_file_handle.close()
      return

    buf = StringIO()
    events_gzip_fd = gzip.GzipFile(mode="w", fileobj=buf)

    read_size = min(CHUNK_SIZE, bytes_to_read)
    events_gzip_fd.write(log_file_handle.read(read_size))
    bytes_to_read -= read_size
    events_gzip_fd.close()
    yield buf.getvalue()

  # Done with this file.
  log_file_handle.close()

def build_gzip_generator_many(file_start_end_tuples, deadline=None):
  """
  For a list of LogFile, start offset, end offset tuples, creates a
  composite generator by calling build_gzip_generator on each file.
  """
  for (log_file, start_offset, end_offset) in file_start_end_tuples:
    log_file_handle = log_file.get_handle()
    for val in build_gzip_generator_single(log_file_handle, start_offset, end_offset, deadline):
      yield val


"""
Wraps the functionality of compressing a set of log files. In general you
should construct this with the paths, types, end time, and total byte limit
you are interested in and then call compress_logs.
"""
class LogCompressor(object):
  def __init__(self, log_path, log_type, end_time, total_byte_limit):
    self._log_path = log_path
    self._log_type = log_type
    self._end_time = end_time
    self._total_byte_limit = total_byte_limit

  def compress_log(self, compress_timeout_millis):
    """
    Compresses the log files associated with this object when it was initialized into
    the passed directory. This may need to handle rolled over log files.
    If total compress time exceeds compress_timeout_millis then incomplete log files 
    will be created.
    This method is a generator for the gzipped data
    """
    deadline = time.time() + int(compress_timeout_millis)/1000.0

    log_files = self.get_log_files()
    files_to_stream = self._prepare_files_for_streaming(log_files, self._log_type, self._end_time, self._total_byte_limit)
    # Scan these from oldest to newest
    return build_gzip_generator_many(files_to_stream, deadline)

  def get_byte_offsets(self, log_file, log_type, end_time, max_size):
    """Returns the byte offsets of the LogFile object log_file ending at end time and
    taking up at most max_size bytes. Returns start_offset, end_offset"""
    if max_size <= 0:
      return 0, 0
    end_offset = log_file.get_end_offset()

    # This may collect very old data up to max_bytes. If end_time is long after
    # the data in the log path ends, we will just collect max_size from the end
    # of the file.
    end_offset = log_search.offset_from_timestamp(end_time, end_offset, log_file.get_path(), log_type)
    start_offset = end_offset - max_size
    if start_offset <= 0:
      return 0, end_offset
    start_offset = self.get_offset_after(log_file.get_path(), log_type, start_offset)
    return start_offset, end_offset

  def _prepare_files_for_streaming(self, log_files, log_type, end_time, byte_limit):
    """Returns a list of LogFile, start offset, end offset tuples to stream. This
    is intended to be used to generate a list of log files that is ready for streaming.
    log_files should be a list of LogFile objects sorted from newest to oldest. 
    The returned list will be orderd from oldest to newest. 
    Total offsets will not exceed total byte limit. 
    The file handles will be open for each log file that should be streamed.
    We need to open the files here in case a rollover occurs during our processing."""
    result = []
    for f in log_files:
      start, end = self.get_byte_offsets(f, log_type, end_time, byte_limit)
      if start != end:
        f.open_log_file()
      else:
        return reversed(result)
      result.append((f, start, end))
      
      # Change the number of bytes that remain
      byte_limit -= end - start
    return reversed(result)
  
  def get_offset_after(self, log_path, log_type, offset):
    """Gets the offset of a log line after a given offset. This is used
    to move to the next entity after a given offset to make sure we don't 
    overrun a maximum size or cut a log line in half.
    If an error occurs or this fails, offset will be returned."""
    f = LogFile(log_path)
    handle = f.open_log_file()

    try:
      event_parser = EVENT_PARSER_CLASSES[log_type]()
      try:
        reader = LogFileEventReader(
          event_parser=event_parser,
          file_handle=handle,
          file_path=log_path,
          start_offset=offset)
        prev_event = reader.find_event_containing_pos(offset)
        event = None
        if prev_event is not None:
          event = reader.get_next_event()
        if event is not None:
          return event.get_offset()
        return offset
      except:
        return offset
    finally:
      handle.close()

  def get_log_files(self):
    """Gets the log files that match log_path* for the log_path this object
    was created with. If more than one file is returned, the newest file will
    be returned first."""
    log_dir, log_name = os.path.split(self._log_path)
    file_pattern_filter = FilePatternFilter(log_name)

    filters = chain.build_chain([DirectoryFilter(), file_pattern_filter])

    if not os.path.exists(log_dir):
      return []

    filter_context = {
      "file_list": [
        LogFile(os.path.join(log_dir, f)) for f in os.listdir(log_dir)
        ]
      }
    filters(filter_context)
    file_list = filter_context["file_list"]
    
    # Now we need to sort these.    
    return sorted(file_list, key=lambda x: self._get_log_file_index(x))
  
  def _get_log_file_index(self, log_file):
    return self.get_log_index(log_file.get_path())

  def get_log_index(self, file_name):
    """Gets an appropriate index for log files where newer files have a smaller
    index. For files that are named 
    x.log, x.log.1, x.log.2... this will be 0, 1, 2... For impala style
    logs like:
    impalad.nightly-5.ent.cloudera.com.impala.log.INFO.20121130-061408.14555
    impalad.nightly-5.ent.cloudera.com.impala.log.INFO.20121130-155821.24443
    ...
    This will be MAX_LONG - time(20121130-061408), MAX_LONG - time(20121130-155821), ..."""
    IMPALA_TIME_FORMAT = "%Y%m%d-%H%M%S"
    MAX_LONG = 2**63 - 1
    
    parts = file_name.split(".")
    if len(parts) > 2:
      try:
        t = time.strptime(parts[-2], IMPALA_TIME_FORMAT)
        return MAX_LONG - time.mktime(t)
      except:
        # Didn't parse. We expect this for non-impala names.
        pass
    try:
      return int(parts[-1])
    except:
      # No number on end of file. Default to 0
      return 0

