// (c) Copyright 2013 Cloudera, Inc. All rights reserved.
define([
  "underscore",
  "knockout",
  "cloudera/Util",
  "cloudera/common/I18n"
], function(_, ko, Util, I18n) {
 
  var healthValues = {
    "RED": 0,
    "YELLOW": 1
  };

  var CheckRow = function(values, entities) {
    var self = this, sample = values[0];

    self.entities = entities;
    self.health = sample.health;
    self.name = sample.name;
    self.number = values.length;

    self.showCheckDetailsPopup = function() {
      $.publish("showCheckDetailsPopup", [self]);
    };
  };

  var EntitySection = function(clusterName, numEntries, entityTypeToEntries,
      entityIdToChecks, displayLimit) {
    var self = this;
    self.clusterName = clusterName;
    self.numEntries = numEntries;

    self.entityTypeToEntries = {};
    self.entityTypeToTotalUnhealthy = {};
    _.each(entityTypeToEntries, function(v, k) {
      if (entityTypeToEntries[k]) {
        self.entityTypeToEntries[k] = v.slice(0, displayLimit);
        self.entityTypeToTotalUnhealthy[k] = v.length;
      }
    });

    self.entityIdToChecks = entityIdToChecks;

    self.notShownString = function(type) {
      var count = entityTypeToEntries[type].length
        - self.entityTypeToEntries[type].length;
      return I18n.t("ui.notShown", count);
    };
  };

  /*
   * options {
   *   updateUri: (required) "the URL to fetch the data"
   *   container: (required) "selector of the container DOM object"
   *   mgmtClusterName: (required) "the name of the mgmt cluster"
   *   displayLimit: (required) "the max number of rows to show per section"
   * }
   */
  var HealthIssues = function(options) {
    var self = this, DISPLAY_LIMIT = options.displayLimit;

    self.loaded = ko.observable(false);

    self.view = ko.observable("ENTITY");
    self.setViewToEntity = function() {
      self.view("ENTITY");
    };
    self.setViewToCheck = function() {
      self.view("CHECK");
    };

    self.unhealthyChecks = ko.observableArray();
    self.totalUnhealthyChecks = ko.observable();
    self.notShownString = ko.computed(function() {
      var count = self.totalUnhealthyChecks() - DISPLAY_LIMIT;
      return I18n.t("ui.notShown", count);
    });
    self.unhealthyEntities = ko.observableArray();

    self.isNoIssues = ko.computed(function() {
      return self.unhealthyChecks().length === 0 &&
        self.unhealthyEntities().length === 0;
    });


    self.showChecks = ko.observable(true);
    self.toggleChecks = function() {
      self.showChecks(!self.showChecks());
    };

    self.showErrors = ko.observable(false);
    self.showConcerning = ko.observable(false);

    self.toggleErrors = function() {
      self.showErrors(!self.showErrors());
    };

    self.toggleConcerning = function() {
      self.showConcerning(!self.showConcerning());
    };

    self.getLevelCount = function(health) {
      var count = 0;
      if (self.view() === "CHECK") {
        count = _.chain(self.unhealthyChecks())
          .groupBy(function(v) {
            return v.name + v.health;
          })
          .filter(function(v, k) {
            return v[0].health === health;
          })
          .value().length;
      } else if (self.view() === "ENTITY") {
        count = _.filter(self.unhealthyEntities(), function(entity) {
          return entity.health === health;
        }).length;
      }
      return count;
    };

    var shouldShowError = function(details) {
      return (self.showErrors() && details.health === 'RED') || details.health !== 'RED';
    };

    var shouldShowConcerning = function(details) {
      return (self.showConcerning() && details.health === 'YELLOW') || details.health !== 'YELLOW';
    };

    /*
     * Function used to sort CheckRow objects.
     *
     * -First sort by health.
     * -If health is the same, sort by number of failing checks,
     *  greatest first.
     * -If number of failing checks is the same, sort
     *  alphabetically
     */
    var checkRowSort = function(a, b) {
      var cmp = healthValues[a.health] - healthValues[b.health];
      if (cmp !== 0) {
        return cmp;
      }

      cmp = b.number - a.number;
      if (cmp !== 0) {
        return cmp;
      }

      return a.name.localeCompare(b.name);
    };

    /*
     * Function used to sort entity objects.
     *
     * -First sort by health.
     * -If health is the same, sort alphabetically.
     */
    var entitySort = function(a, b) {
      var cmp = healthValues[a.health] - healthValues[b.health];
      if (cmp !== 0) {
        return cmp;
      }

      return a.name.localeCompare(b.name);
    };

    self.unhealthyCheckRows = ko.computed(function() {
      var checkRows = _.chain(self.unhealthyChecks())
        .filter(shouldShowError)
        .filter(shouldShowConcerning)
        .groupBy(function(check) {
          return check.name + check.health;
        })
        .map(function(v, k) {
          var entities = _.filter(self.unhealthyEntities(), function(entity) {
            var match = !_.every(v, function(check) {
              return check.entityId !== entity.entityId;
            });

            return match;
          });
          return new CheckRow(v, entities);
        })
      .value();

      checkRows = checkRows.sort(checkRowSort);
      self.totalUnhealthyChecks(checkRows.length);
      return checkRows.slice(0, DISPLAY_LIMIT);
    });

    /*
     * Function to sort EntitySection objects.  Sorts alphabetically
     * by cluster name, but always places the MGMT cluster at the
     * end.
     */
    var entitySectionSort = function(a, b) {
      if (a.clusterName === options.mgmtClusterName) {
        return 1;
      } else if (b.clusterName === options.mgmtClusterName) {
        return -1;
      } else {
        return a.clusterName.localeCompare(b.clusterName);
      }
    };

    self.unhealthyEntitySections = ko.computed(function() {
      var entitySections = _.chain(self.unhealthyEntities())
        .filter(shouldShowError)
        .filter(shouldShowConcerning)
        .groupBy(function(entity) {
          return entity.clusterName;
        })
        .map(function(entities, clusterName) {
          var entityIdToChecks = {};
          _.each(entities, function(entity) {
            var failingChecks = _.chain(self.unhealthyChecks())
              .filter(function(check) {
                return entity.entityId === check.entityId;
              })
              .map(function(check) {
                return check.name;
              })
            .value();

            entityIdToChecks[entity.entityId] = failingChecks.join(", ");
          });

          // map of entityType (SERVICE, ROLE, or HOST) to entities of
          // that type
          var entityTypeToEntries = {};
          _.each(_.groupBy(entities, "entityType"), function(v, k) {
            entityTypeToEntries[k] = v.sort(entitySort);
          });

          return new EntitySection(clusterName, entities.length, entityTypeToEntries,
            entityIdToChecks, DISPLAY_LIMIT);
        })
      .value();

      return entitySections.sort(entitySectionSort);
    });

    self.isNoResults = ko.computed(function() {
      if (self.view() === "CHECK") {
        return self.unhealthyCheckRows().length === 0;
      } else if (self.view() === "ENTITY") {
        return self.unhealthyEntitySections().length === 0;
      } else {
        return true;
      }
    });

    self.hiddenClusters = ko.observableArray();
    self.allVisibleClusters = ko.computed(function() {
      return _.pluck(self.unhealthyEntitySections(), "clusterName");
    });

    self.isClusterCollapsed = function(clusterName) {
      return _.indexOf(self.hiddenClusters(), clusterName) !== -1;
    };

    self.toggleCluster = function(clusterName) {
      if (self.hiddenClusters.remove(clusterName).length === 0) {
        self.hiddenClusters.push(clusterName);
      }
    };

    /*
     * Collapses all Health Issues subsections.
     */
    self.collapseAll = function() {
      if (self.view() === "CHECK") {
        self.showChecks(false);
      } else if (self.view() === "ENTITY") {
        self.hiddenClusters.removeAll();
        _.each(self.allVisibleClusters(), function(clusterName) {
          self.hiddenClusters.push(clusterName);
        });
      }
    };

    /*
     * Expands all Health Issues subsections.
     */
    self.expandAll = function() {
      if (self.view() === "CHECK") {
        self.showChecks(true);
      } else if (self.view() === "ENTITY") {
        self.hiddenClusters.removeAll();
      }
    };

    /*
     * Returns true if any subsections are open. In this
     * case we should show an "Expand All" link.
     *
     * Returns false if all subsections are closed.  In this
     * case we should show a "Collapse All" link.
     */
    self.showExpandAll = ko.computed(function() {
      if (self.view() === "CHECK") {
        return !self.showChecks();
      } else if (self.view() === "ENTITY") {
        return self.hiddenClusters().length === self.allVisibleClusters().length;
      } else {
        return false;
      }
    });

    var updateData = function(response) {
      self.unhealthyChecks.removeAll();
      _.each(response.unhealthyChecks, function(check) {
        self.unhealthyChecks.push(check);
      });

      self.unhealthyEntities.removeAll();
      _.each(response.unhealthyEntities, function(entity) {
        self.unhealthyEntities.push(entity);
      });
    };

    var initData = function(response) {
      self.loaded(true);

      updateData(response);

      self.showErrors(self.getLevelCount("RED") > 0);
      self.showConcerning(!self.showErrors() && self.getLevelCount("YELLOW") > 0);
    };

    var handleError = function(response) {
      Util.filterError(response.responseText);
    };

    var loadData = function(now, currentMode) {
      $.post(options.updateUri, 'json')
        .success(updateData)
        .error(handleError);
    };

    $.post(options.updateUri, 'json')
      .success(initData)
      .error(handleError);

    var handle = $.subscribe("markerDateChanged", loadData);
    self.subscriptionHandles = [handle];
    self.unsubscribe = function() {
      Util.unsubscribe(self);
    };

    // Add bootstrap tooltips to filters.
    $(options.container).find(".RegionTitle button.showTooltip").tooltip();
  };

  return HealthIssues;
});
