// (c) Copyright 2011-2012 Cloudera, Inc. All rights reserved.
define([
  'cloudera/search/SearchResults',
  'cloudera/search/FilterModel',
  'knockout'
], function(SearchResults, FilterModel, ko) {

  describe('SearchResults', function() {

    var $resultList, dataSourceSpy,
      searchResults, goodResponse;

    beforeEach(function() {
      jasmine.Ajax.useMock();
      // Dummy success JSON to return from mock responses.
      goodResponse = {
        status: 200,
        responseText: JSON.stringify({success: true})
      };
      // Generate the mock data that we're sending in to the result
      // search instance.
      $resultList = $('<div></div>').appendTo($('body'));
      dataSourceSpy = jasmine.createSpy();

      searchResults = new SearchResults($resultList[0], dataSourceSpy);
    });

    afterEach(function() {
      $resultList.remove();
    });

    it('defines some attributes', function() {
      expect(searchResults.error).toBeDefined();
      expect(searchResults.results).toBeDefined();
      expect(searchResults.loading).toBeDefined();
      expect(searchResults.dataSource).toBeDefined();
    });

    it('applies bindings upon instantiation', function() {
      spyOn(ko, 'applyBindings');
      searchResults = new SearchResults($resultList[0], dataSourceSpy);
      expect(ko.applyBindings.callCount).toEqual(1);
      var args = ko.applyBindings.argsForCall[0];
      expect(args[1]).toEqual($resultList[0]);
    });

    it('exposes a subset of itself as a view model to the result list', function() {
      spyOn(ko, 'applyBindings');
      searchResults = new SearchResults($resultList[0], dataSourceSpy);
      expect(ko.applyBindings.callCount).toEqual(1);
      var args = ko.applyBindings.argsForCall[0];
      var viewModel = args[0];
      expect(viewModel.loading).toEqual(searchResults.loading);
      expect(viewModel.error).toEqual(searchResults.error);
      expect(viewModel.results).toEqual(searchResults.results);
    });

    it('exposes hook for modifying view model passed to applyBindings', function() {
      var modifyViewModel = jasmine.createSpy('modifyViewModel');
      searchResults = new SearchResults($resultList[0], dataSourceSpy,
        modifyViewModel);
      expect(modifyViewModel).wasCalled();
      // Were we given the view model?
      var viewModel = modifyViewModel.mostRecentCall.args[0];
      expect(viewModel.loading).toBeDefined();
      expect(viewModel.error).toBeDefined();
      expect(viewModel.results).toBeDefined();
    });

    it('getHead and getTail work when there are no results', function() {
      expect(searchResults.length()).toEqual(0);
      expect(searchResults.getHead()).toBeNull();
      expect(searchResults.getTail()).toBeNull();
    });

    it('getHead and getTail work when there are results', function() {
      searchResults.results([1, 2, 3]);
      expect(searchResults.length()).toEqual(3);
      expect(searchResults.getHead()).toEqual(1);
      expect(searchResults.getTail()).toEqual(3);
    });

    it('can splice results to the end', function() {
      searchResults.results([1, 2, 3]);
      expect(searchResults.length()).toEqual(3);
      searchResults.splice(0);
      expect(searchResults.length()).toEqual(0);
      searchResults.results([1, 2, 3]);
      searchResults.splice(3, 0, 4);
      expect(searchResults.length()).toEqual(4);
      var results = searchResults.results();
      expect(results[0]).toEqual(1);
      expect(results[1]).toEqual(2);
      expect(results[2]).toEqual(3);
      expect(results[3]).toEqual(4);
    });

    it('correctly prepends results', function() {
      // Test the single result case.
      var result1 = new FilterModel();
      result1.name = 'result1';
      var result = searchResults.prepend(result1);
      expect(result).toEqual(1);

      // Test the array case.
      var result2 = new FilterModel();
      result2.name = 'result2';
      var result3 = new FilterModel();
      result3.name = 'result3';
      result = searchResults.prepend([result2, result3]);
      expect(result).toEqual(3);
      var results = searchResults.results();
      expect(results[0].name).toEqual('result2');
      expect(results[1].name).toEqual('result3');
      expect(results[2].name).toEqual('result1');

      // Test the prepend to empty results case.
      searchResults.results([]);
      expect(searchResults.length()).toEqual(0);
      result = searchResults.prepend([result1, result2, result3]);
      expect(result).toEqual(searchResults.length());
    });

    it('correctly appends results', function() {
      // Test the single result case.
      var result1 = new FilterModel();
      result1.name = 'result1';
      var result = searchResults.append(result1);
      expect(result).toEqual(1);

      // Test the array case.
      var result2 = new FilterModel();
      result2.name = 'result2';
      var result3 = new FilterModel();
      result3.name = 'result3';
      result = searchResults.append([result2, result3]);
      expect(result).toEqual(3);
      var results = searchResults.results();
      expect(results[0].name).toEqual('result1');
      expect(results[1].name).toEqual('result2');
      expect(results[2].name).toEqual('result3');

      // Test the append to empty results case.
      searchResults.results([]);
      expect(searchResults.length()).toEqual(0);
      result = searchResults.append([result1, result2, result3]);
      expect(result).toEqual(searchResults.length());
    });

    it('has method for retrying the dataSource after a time', function() {
      var clear = true;
      var handler = jasmine.createSpy();
      var timeoutInMs = 5;
      runs(function() {
        searchResults.retryHandlerAfterTime(clear, handler, timeoutInMs);
      });
      waitsFor(function() {
        return dataSourceSpy.callCount > 0;
      }, 'dataSource never called', timeoutInMs * 2);
      runs(function() {
        // Verify that dataSource was called with correct args.
        var args = dataSourceSpy.mostRecentCall.args;
        expect(args[0]).toEqual(searchResults);
        expect(args[1]).toEqual(clear);
        expect(args[2]).toEqual(handler);
      });
    });

    it('knows how to handle errors when loading data', function() {
      // We're not providing a callback here so we don't have to
      // worry about async retries.
      var handler = searchResults.makeLoadDataHandler(true);
      expect(searchResults.error()).toBeFalsy();
      searchResults.loading(true);

      handler(null);
      expect(searchResults.error()).toBeFalsy();
      expect(searchResults.loading()).toBeFalsy();

      searchResults.loading(true);
      var errorMessage = 'catpants';
      handler(errorMessage);
      expect(searchResults.error()).toEqual(errorMessage);
      expect(searchResults.loading()).toBeFalsy();
    });

    it('knows how to call callbacks when loading data', function() {
      var callback = jasmine.createSpy().andReturn(false);
      spyOn(searchResults, 'retryHandlerAfterTime');
      var handler = searchResults.makeLoadDataHandler(true, callback);

      handler(null);
      expect(callback).wasCalled();
      expect(searchResults.retryHandlerAfterTime).wasNotCalled();

      callback.andReturn(true);
      handler('error');
      expect(callback).wasCalledWith('error');
      expect(searchResults.retryHandlerAfterTime).wasCalled();
    });

    it('knows how to load data', function() {
      var handler = 'handler';
      var callback = function() {};
      spyOn(searchResults, 'makeLoadDataHandler').andReturn(handler);
      searchResults.error('catpants error');
      expect(searchResults.loading()).toBeFalsy();

      searchResults.loadData(true, callback);
      expect(searchResults.error()).toBeFalsy();
      expect(searchResults.loading()).toBeTruthy();
      expect(dataSourceSpy).wasCalledWith(searchResults, true, handler);

      searchResults.error('catpants error');
      searchResults.loading(false);
      dataSourceSpy.reset();
      searchResults.loadData(callback);
      expect(searchResults.error()).toBeFalsy();
      expect(searchResults.loading()).toBeTruthy();
      expect(dataSourceSpy).wasCalledWith(searchResults, undefined, handler);
    });

    it('can clear results', function() {
      searchResults.results([1, 2, 3]);
      searchResults.clear();
      expect(searchResults.results().length).toEqual(0);
    });
  });
});